#!/usr/bin/python

import os
import time
import datetime
import sys
import subprocess

from mininet.topo import Topo
from mininet.node import Controller
from mininet.node import RemoteController
from mininet.node import OVSController
from mininet.log import info, error, debug, output, warn, setLogLevel
from mininet.examples.cluster import MininetCluster, SwitchBinPlacer, RoundRobinPlacer
from mininet.examples.clustercli import ClusterCLI as CLI

from sgcommmodel import *
from sgmininettopology import *
from commandgenerator import CommandGenerator
from nfcollector import *
from sample_test import *

class SGPerformanceEvaluationDriver:
    """
	SG Test represents the executes the test specified with SimulationParameter
	and SGCommunicationTopology.

	It is responsible to initialize mininet network, limit link bandwidt, start
	defined flows and start NetFlow collectors.
	"""
    _model = None
    _parameters = None
    _nf_collector = None
    _switch_to_host = None
    net = None

    # Environment paths
    _output_path = "~/logs"
    _simulators_path="~/sm_simulator"

    def __init__(self, model, parameters):
        """
        Constructor that saves model and parameters and creates NFCollector.
        
        Arguments:
        model - SGCommunicationModel object
        parameters - SimulationParameters object specifying test paremeters.
        """
        self._model = model
        self._parameters = parameters
        self._nf_collector = NFCollector(self._parameters.nodes_path, self._output_path)

    def initialize_network(self):
        """
        Creates MininetCluster object.
        """
        print self._parameters.nodes_path 
        servers = self._get_nodes(self._parameters.nodes_path)
        self.net, self._switch_to_host = CustomNet.GenerateNet_2(self._model.topology, servers)

    def configure_network(self):
        """
        Configures the communication network for all the features
        that require running mininet (limiting switch links netflow
        providers and netflow collectors).
        """    
        self._limit_egress_rate()
        self._limit_ingress_rate()
        self._initialize_nflow_providers()
        self._initialize_netflow_collectors()

    def execute_test(self):
        """
        Execute tests by starting configured flows and waiting for
        paremeterized duration.
        """
        self._start_flows()        
        for iteration in range(0, self._parameters.duration):
            if iteration % 30 == 0:
                print "Waiting: " + str(iteration) + "\\" + str(self._parameters.duration)
            time.sleep(1)

        print "Executed tests."

    def deinitialize(self):
        """
        Stops the emulation environment.
        """
        self.net.stop()
        self._deinitialize_netflow_collectors()

    # Private API
    def _start_mininet(self):
        self.net.start()

    def _start_flows(self):
        if self._model.topology.flows:
            for flow in self._model.topology.flows:
                source = flow.source
                destination = flow.destination

                source_host = self._switch_to_host[source[0].name]
                port = source[1]
                destination_host = self._switch_to_host[destination[0].name]

                info(source_host, ":", str(port), "<->", destination_host, '\n')

                srcNode = self.net.getNodeByName(source_host.name)
                dstNode = self.net.getNodeByName(destination_host.name)

                command = flow.generate_server_command(self._simulators_path, self._output_path)
                print srcNode.name + " " + srcNode.cmd(command)
                
                command = flow.generate_client_command(self._simulators_path, self._output_path, srcNode.IP())
                print dstNode.name + " " + dstNode.cmd(command)

    # Objasniti zbog cega ne moze da se koristi tc
    def _limit_egress_rate(self):
        for switch in self._model.topology.Switches():
            mininet_switch = self.net.getNodeByName(switch.name)
            bw = switch.bw
            self._limit_egress_rate_for_switch(mininet_switch, bw)

    def _limit_ingress_rate(self):
        for switch in self._model.topology.Switches():
            mininet_switch = self.net.getNodeByName(switch.name)
            bw = switch.bw
            self._limit_ingress_rate_for_switch(mininet_switch, bw)

    def _limit_egress_rate_for_switch(self, switch, maxRate):
        # loop trough all interfaces skipping the first one (lo)
        for intName in switch.intfNames()[1:]:
            command = CommandGenerator.queue_command(intName, maxRate*1000*1000)
            #print switch.name + ': ' + command
            print switch.cmd(command)
    
    def _limit_ingress_rate_for_switch(self, switch, maxRate):
        # loop trough all interfaces skipping the first one (lo)
        for intName in switch.intfNames()[1:]:
            command = CommandGenerator.policing_ingress_rate_command(intName, maxRate*1000*1000)
            switch.cmd(command)

            command = CommandGenerator.policing_ingress_burst_command(intName, 2 * 1024)
            switch.cmd(command)

    def _initialize_nflow_providers(self):
        for switch in self.net.switches:
            command = CommandGenerator.clear_provider_command(switch.name)
            switch.cmdPrint(command)

            command = CommandGenerator.create_provider_command(switch.name)
            switch.cmdPrint(command)

    def _initialize_netflow_collectors(self):
        if self._parameters.use_netflow:
            self._nf_collector.start_collectors()

    def _deinitialize_netflow_collectors(self):
        if self._parameters.use_netflow:
            self._nf_collector.stop_collectors()

    def _get_nodes(self, nodes_path):
        """
        Parse nodes file and remove trailing newline before returning
        nodes list.
        """
        nodes_ = tuple(open(nodes_path, 'r'))
        nodes = []
        for node in nodes_:
            nodes.append(node.strip())

        return nodes

def main():
    "ADMS Data Agregation simulation"

    parameters = SampleTestCase.parameters()
    custom_topology = SampleTestCase.topology()

    model = SGCommunicationModel(custom_topology, parameters)

    driver = SGPerformanceEvaluationDriver(model, model.parameters)

    driver.initialize_network()

    driver.configure_network()
    
    driver.execute_test()

    driver.deinitialize()
    
    print "ADMS Data Agregation simulation ended."


def usage():
    print "Usage:"
    print "sudo -E ./SG-TestBench.py"

if __name__ == '__main__':
    #setLogLevel( 'debug' )
    setLogLevel( 'info' )
    main()


